@@ -27,6 +27,7 @@
 #include <sys/types.h>
 #include <pwd.h>
 #include <grp.h>
+#include <unistd.h>
 #include <sstream>
 #include <locale>
 #include <string>
@@ -111,15 +112,50 @@ namespace snapper
     };
 
 
-    string sformat(const string& format, ...);
+    struct FdCloser
+    {
+	FdCloser(int fd)
+	    : fd(fd)
+	{
+	}
+
+	~FdCloser()
+	{
+	    if (fd > -1 )
+		::close(fd);
+	}
+
+	void reset()
+	{
+	    fd = -1;
+	}
+
+	int close()
+	{
+	    int r = ::close(fd);
+	    fd = -1;
+	    return r;
+	}
+
+    private:
+
+	int fd;
+
+    };
+
+
+    string sformat(const char* format, ...) __attribute__ ((format(printf, 1, 2)));
 
 
     struct runtime_error_with_errno : public std::runtime_error
     {
-	explicit runtime_error_with_errno(const char* what_arg, int errnum)
-	    : runtime_error(sformat("%s, errno:%d (%s)", what_arg, errnum,
-				    stringerror(errnum).c_str()))
+	explicit runtime_error_with_errno(const char* what_arg, int error_number)
+	    : runtime_error(sformat("%s, errno:%d (%s)", what_arg, error_number,
+				    stringerror(error_number).c_str())),
+	      error_number(error_number)
 	{}
+
+	const int error_number;
     };
 
 }
