@@ -1,5 +1,6 @@
 /*
  * Copyright (c) [2011-2014] Novell, Inc.
+ * Copyright (c) [2018-2020] SUSE LLC
  *
  * All Rights Reserved.
  *
@@ -34,12 +35,18 @@
 #include <errno.h>
 #include <stdlib.h>
 #include <assert.h>
+#ifdef ENABLE_SELINUX
+#include <selinux/selinux.h>
+#endif
 #include <algorithm>
 
 #include "snapper/FileUtils.h"
 #include "snapper/AppUtil.h"
 #include "snapper/Log.h"
 #include "snapper/Exception.h"
+#ifdef ENABLE_SELINUX
+#include "snapper/Selinux.h"
+#endif
 
 
 namespace snapper
@@ -55,23 +62,16 @@ namespace snapper
     {
 	dirfd = ::open(base_path.c_str(), O_RDONLY | O_NOATIME | O_CLOEXEC);
 	if (dirfd < 0)
-	{
-	    y2err("open failed path:" << base_path << " error:" << stringerror(errno));
-	    throw IOErrorException();
-	}
+	    SN_THROW(IOErrorException(sformat("open failed path:%s errno:%d (%s)", base_path.c_str(),
+					      errno, stringerror(errno).c_str())));
 
 	struct stat buf;
 	if (fstat(dirfd, &buf) != 0)
-	{
-	    y2err("fstat failed path:" << base_path << " error:" << stringerror(errno));
-	    throw IOErrorException();
-	}
+	    SN_THROW(IOErrorException(sformat("fstat failed path:%s errno:%d (%s)", base_path.c_str(),
+					      errno, stringerror(errno).c_str())));
 
 	if (!S_ISDIR(buf.st_mode))
-	{
-	    y2err("not a directory path:" << base_path);
-	    throw IOErrorException();
-	}
+	    SN_THROW(IOErrorException("not a directory path:" + base_path));
 
 	setXaStatus();
     }
@@ -86,22 +86,20 @@ namespace snapper
 	dirfd = ::openat(dir.dirfd, name.c_str(), O_RDONLY | O_NOFOLLOW | O_NOATIME | O_CLOEXEC);
 	if (dirfd < 0)
 	{
-	    y2err("open failed path:" << dir.fullname(name) << " (" << stringerror(errno) << ")");
-	    throw IOErrorException();
+	    string failed_path = dir.fullname() + "/" + name;
+	    SN_THROW(IOErrorException(sformat("open failed path:%s errno:%d (%s)", failed_path.c_str(),
+					      errno, stringerror(errno).c_str())));
 	}
 
 	struct stat buf;
 	if (fstat(dirfd, &buf) != 0)
-	{
-	    y2err("fstat failed path:" << base_path << " error:" << stringerror(errno));
-	    throw IOErrorException();
-	}
+	    SN_THROW(IOErrorException(sformat("fstat failed path:%s errno:%d (%s)", base_path.c_str(),
+					      errno, stringerror(errno).c_str())));
 
 	if (!S_ISDIR(buf.st_mode))
 	{
-	    y2err("not a directory path:" << dir.fullname(name));
 	    close(dirfd);
-	    throw IOErrorException();
+	    SN_THROW(IOErrorException("not a directory path:" + dir.fullname(name)));
 	}
 
 	xastatus = dir.xastatus;
@@ -113,10 +111,8 @@ namespace snapper
     {
 	dirfd = fcntl(dir.dirfd, F_DUPFD_CLOEXEC, 0);
 	if (dirfd == -1)
-	{
-	    y2err("fcntl(F_DUPFD_CLOEXEC) failed error:" << stringerror(errno));
-	    throw IOErrorException();
-	}
+	    SN_THROW(IOErrorException(sformat("fcntl(F_DUPFD_CLOEXEC) failed error:%d (%s)", errno,
+					      stringerror(errno).c_str())));
 
 	xastatus = dir.xastatus;
     }
@@ -130,10 +126,8 @@ namespace snapper
 	    ::close(dirfd);
 	    dirfd = fcntl(dir.dirfd, F_DUPFD_CLOEXEC, 0);
 	    if (dirfd == -1)
-	    {
-		y2err("fcntl(F_DUPFD_CLOEXEC) failed error:" << stringerror(errno));
-		throw IOErrorException();
-	    }
+		SN_THROW(IOErrorException(sformat("fcntl(F_DUPFD_CLOEXEC) failed error:%d (%s)", errno,
+						  stringerror(errno).c_str())));
 
 	    xastatus = dir.xastatus;
 	}
@@ -192,22 +186,41 @@ namespace snapper
     {
 	int fd = fcntl(dirfd, F_DUPFD_CLOEXEC, 0);
 	if (fd == -1)
-	{
-	    y2err("fcntl(F_DUPFD_CLOEXEC) failed error:" << stringerror(errno));
-	    throw IOErrorException();
-	}
+	    SN_THROW(IOErrorException(sformat("fcntl(F_DUPFD_CLOEXEC) failed error:%d (%s)", errno,
+					      stringerror(errno).c_str())));
 
 	DIR* dp = fdopendir(fd);
 	if (dp == NULL)
 	{
-	    y2err("fdopendir failed path:" << fullname() << " error:" << stringerror(errno));
 	    ::close(fd);
-	    throw IOErrorException();
+	    SN_THROW(IOErrorException(sformat("fdopendir failed path:%s error:%d (%s)",
+					      fullname().c_str(), errno, stringerror(errno).c_str())));
 	}
 
 	vector<string> ret;
 
-	size_t len = offsetof(struct dirent, d_name) + fpathconf(dirfd, _PC_NAME_MAX) + 1;
+#if defined(__GLIBC__) && ((__GLIBC__ > 2) || ((__GLIBC__ == 2) && (__GLIBC_MINOR__ >= 24)))
+
+	// Since glibc 2.24 readdir is thread safe under certain
+	// condidtions, which apply here, and readdir_r is deprecated
+	// (see readdir(3)).
+
+	struct dirent* ep = nullptr;
+
+	rewinddir(dp);
+	while ((ep = readdir(dp)) != nullptr)
+	{
+	    if (strcmp(ep->d_name, ".") != 0 && strcmp(ep->d_name, "..") != 0 &&
+		pred(ep->d_type, ep->d_name))
+		ret.push_back(ep->d_name);
+	}
+
+#else
+
+	long sz = fpathconf(dirfd, _PC_NAME_MAX);
+	if (sz == -1)
+	    sz = NAME_MAX;
+	size_t len = offsetof(struct dirent, d_name) + sz + 1;
 	struct dirent* ep = (struct dirent*) malloc(len);
 	struct dirent* epp;
 
@@ -221,6 +234,8 @@ namespace snapper
 
 	free(ep);
 
+#endif
+
 	closedir(dp);
 
 	return ret;
@@ -364,6 +379,13 @@ namespace snapper
     }
 
 
+    int
+    SDir::fsync() const
+    {
+	return ::fsync(dirfd);
+    }
+
+
     int
     SDir::mktemp(string& name) const
     {
@@ -514,9 +536,9 @@ namespace snapper
 	    }
 	    else
 	    {
-                y2err("Couldn't get extended attributes status for " << base_path << "/" <<
-		      path << stringerror(errno));
-                throw IOErrorException();
+                SN_THROW(IOErrorException(sformat("Couldn't get extended attributes status for %s/%s, "
+						  "errno:%d (%s)", base_path.c_str(), path.c_str(),
+						  errno, stringerror(errno).c_str())));
 	    }
 	}
 	else
@@ -582,6 +604,171 @@ namespace snapper
     }
 
 
+    bool
+    SDir::fsetfilecon(const string& name, char* con) const
+    {
+	assert(name.find('/') == string::npos);
+	assert(name != "..");
+
+	bool retval = true;
+
+#ifdef ENABLE_SELINUX
+	if (_is_selinux_enabled())
+	{
+	    char *src_con = NULL;
+
+	    int fd = ::openat(dirfd, name.c_str(), O_RDONLY | O_NOFOLLOW | O_NOATIME
+			      | O_NONBLOCK | O_CLOEXEC);
+	    if (fd < 0)
+	    {
+		// symlink, detached dev node?
+		if (errno != ELOOP && errno != ENXIO && errno != EWOULDBLOCK)
+		{
+		    y2err("open failed errno: " << errno << " (" << stringerror(errno) << ")");
+		    return false;
+		}
+
+		boost::lock_guard<boost::mutex> lock(cwd_mutex);
+
+		if (fchdir(dirfd) < 0)
+		{
+		    y2err("fchdir failed errno: " << errno << " (" << stringerror(errno) << ")");
+		    return false;
+		}
+
+		if (lgetfilecon(name.c_str(), &src_con) < 0 || selinux_file_context_cmp(src_con, con))
+		{
+		    y2deb("setting new SELinux context on " << fullname() << "/" << name);
+		    if (lsetfilecon(name.c_str(), con))
+		    {
+			y2err("lsetfilecon on " << fullname() << "/" << name << " failed errno: " << errno << " (" << stringerror(errno) << ")");
+			retval = false;
+		    }
+		}
+
+		chdir("/");
+
+	    }
+	    else
+	    {
+		if (fgetfilecon(fd, &src_con) < 0 || selinux_file_context_cmp(src_con, con))
+		{
+		    y2deb("setting new SELinux context on " << fullname() << "/" << name);
+		    if (::fsetfilecon(fd, con))
+		    {
+			y2err("fsetfilecon on " << fullname() << "/" << name << " failed errno: " << errno << " (" << stringerror(errno) << ")");
+			retval = false;
+		    }
+		}
+
+		::close(fd);
+	    }
+
+	    freecon(src_con);
+	}
+#endif
+	return retval;
+    }
+
+
+    bool
+    SDir::restorecon(const string& name, SelinuxLabelHandle* sh) const
+    {
+	assert(name.find('/') == string::npos);
+	assert(name != "..");
+
+	bool retval = true;
+#ifdef ENABLE_SELINUX
+	if (_is_selinux_enabled())
+	{
+	    assert(sh);
+
+	    struct stat buf;
+	    if (stat(name, &buf, AT_SYMLINK_NOFOLLOW))
+	    {
+		y2err("Failed to stat " << fullname() << "/" << name);
+		return false;
+	    }
+
+	    char* con = sh->selabel_lookup(fullname() + "/" + name, buf.st_mode);
+	    if (con)
+	    {
+		retval = fsetfilecon(name, con);
+	    }
+	    else
+	    {
+		retval = false;
+	    }
+
+	    freecon(con);
+	}
+#endif
+	return retval;
+    }
+
+
+    bool
+    SDir::fsetfilecon(char* con) const
+    {
+	bool retval = true;
+
+#ifdef ENABLE_SELINUX
+	if (_is_selinux_enabled())
+	{
+	    char* src_con = NULL;
+
+	    if (fgetfilecon(fd(), &src_con) < 0 || selinux_file_context_cmp(src_con, con))
+	    {
+		y2deb("setting new SELinux context on " << fullname());
+		if (::fsetfilecon(fd(), con))
+		{
+		    y2err("fsetfilecon on " << fullname() << " failed errno: " << errno << " (" << stringerror(errno) << ")");
+		    retval = false;
+		}
+	    }
+
+	    freecon(src_con);
+	}
+#endif
+	return retval;
+    }
+
+
+    bool
+    SDir::restorecon(SelinuxLabelHandle* sh) const
+    {
+	bool retval = true;
+#ifdef ENABLE_SELINUX
+	if (_is_selinux_enabled())
+	{
+	    assert(sh);
+
+	    struct stat buf;
+
+	    if (stat(&buf))
+	    {
+		y2err("Failed to stat " << fullname());
+		return false;
+	    }
+
+	    char* con = sh->selabel_lookup(fullname(), buf.st_mode);
+	    if (con)
+	    {
+		retval = fsetfilecon(con);
+	    }
+	    else
+	    {
+		y2war("can't get proper label for path:" << fullname());
+		retval = false;
+	    }
+
+	    freecon(con);
+	}
+#endif
+	return retval;
+    }
+
+
     SFile::SFile(const SDir& dir, const string& name)
 	: dir(dir), name(name)
     {
@@ -646,6 +833,19 @@ namespace snapper
     }
 
 
+    void
+    SFile::fsetfilecon(char* con) const
+    {
+	dir.fsetfilecon(name, con);
+    }
+
+    void
+    SFile::restorecon(SelinuxLabelHandle* sh) const
+    {
+	dir.restorecon(name, sh);
+    }
+
+
     TmpDir::TmpDir(SDir& base_dir, const string& name_template)
 	: base_dir(base_dir), name(name_template)
     {
